;;;
;;; pserializer extension sample
;;;
;;;   Copyright (c) 1999 Shiro Kawai (shiro@acm.org)
;;;   Permission to use, modify, and distribute of this code is granted
;;;   under the same condition as of STk.
;;;
;;; $Id: example.scm,v 1.1 2000/01/02 05:47:34 shiro Exp $
;;;


;; This file shows how to extend the pserializer module.
;;
;; STk has a native support of hashtable.  MAKE-HASH-TABLE creates
;; a hash table with eq? comparison function and default hash function.
;; This code allows to serializer this type of hashtables.
;;
;; Although you can use different comparison and hash functions to
;; create hash tables in STk, there's no interface provided to query
;; those functions from existing hash tables.  Thus, deserializer
;; assumes the hashtable uses default comparison and hash functions.

(require "pserializer")

(define *tag-hashtable* 'ht)            ; choose 'ht as a tag.

(define (test-hashtable obj)
  (hash-table? obj))

(define (write-hashtable obj serializer)
  (let ((num 0))
    ;; put # of entries
    (hash-table-for-each obj (lambda _ (set! num (+ num 1))))
    (write-to-output-serializer num serializer)
    ;; put keys and values
    (hash-table-for-each
     obj
     (lambda (key value)
       (write-to-output-serializer key serializer)
       (write-to-output-serializer value serializer)))))

(define (read-hashtable serializer)
  (let ((num (read-from-input-serializer serializer))
        (ht  (make-hash-table)))
    ;; Note: register the hashtable being read to the serializer here,
    ;; or the reference counter won't be in sync.
    (register-object-to-input-serializer ht serializer)
    ;; Read keys and values.
    (do ((num num (- num 1)))
        ((= num 0) ht)
      (let* ((key (read-from-input-serializer serializer))
             (val (read-from-input-serializer serializer)))
        (hash-table-put! ht key val)))))

(define *serializer-extension*
  `((,*tag-hashtable* ,test-hashtable ,write-hashtable ,read-hashtable)))


;; test 

(define h (make-hash-table))

(hash-table-put! h 'a 324)
(hash-table-put! h 'b '(a b c d e))

; sharing the same value
(define x '(x y z))
(hash-table-put! h 'c x)
(hash-table-put! h 'd x)

; serialize->deserialize

(define serialized
  (call-with-output-string
   (lambda (port)
     (call-with-output-serializer
      port
      (lambda (serializer) (write-to-output-serializer h serializer))
      *serializer-extension*))))

(define h2
  (call-with-input-string serialized
   (lambda (port)
     (call-with-input-serializer
      port
      (lambda (serializer) (read-from-input-serializer serializer))
      *serializer-extension*))))

; see if it's ok

(if (and (equal? (hash-table-get h 'a) (hash-table-get h2 'a))
         (equal? (hash-table-get h 'b) (hash-table-get h2 'b))
         (equal? (hash-table-get h 'c) (hash-table-get h2 'c))
         (eq? (hash-table-get h2 'c) (hash-table-get h2 'd)))
    (display "Ok.")
    (display "Failed."))
(newline)

