;;;
;;; pserializer.scm - 'portable' serializer
;;;
;;;   Copyright (c) 1999 Shiro Kawai (shiro@acm.org)
;;;   Permission to use, modify, and distribute of this code is granted
;;;   under the same condition as of STk.
;;;
;;; $Id: pserializer.scm,v 1.2 2000/01/25 12:43:59 shiro Exp $
;;;

;; This is a simple implementation to serialize standard Scheme
;; objects (boolean, pair, symbol, string, number, character, string
;; and vector).   Serialization is a function to convert a Scheme
;; structure to a certain bytestream which is independent from the
;; running process, and when it is read back ("deserialized"), it
;; recovers a Scheme structure topologically equal to the original one.
;; Serialized form is very useful to store a Scheme structure in a
;; file (persistence), or to send it over the network.
;;
;; This implementation intended to be portable and extensible
;; for various Scheme implementations.
;;
;; To use this module, you need to define a few hash-table funcitons
;; and error function.  The example to use SLIB and STk is shown below.
;; At least SLIB version should work on most Scheme systems.
;; See "Implementation dependent part" below.
;;
;; Although this implementation deals with only the standard Scheme
;; objects, you can extend the serializer routine to accept other
;; types of objects specific to your Scheme implementation (e.g. records
;; or classes).
;; 
;; External Interface
;;
;;   MAKE-OUTPUT-SERIALIZER port &optional extension      [function]
;;
;;     Create a serializer from given output port PORT and returns
;;     it.  Optional argument WRITER-EXTENSION is used to extend
;;     the serializer to recognize implementation dependent objects.
;;
;;   WRITE-TO-OUTPUT-SERIALIZER object serializer         [function]
;;
;;     Write an OBJECT to the output serializer SERIALIZER.
;;
;;   CALL-WITH-OUTPUT-SERIALIZER port proc &optional extension   [function]
;;
;;     PROC must be a procedure which takes one argument, serializer.
;;     This procedure creates a serializer from PORT and
;;     WRITER-EXTENSION argument, and passes it to PROC.
;;
;;   MAKE-INPUT-SERIALIZER port &optional extension       [function]
;;
;;     Create a deserializer from given input port PORT and returns
;;     it.  Optional argument READER-EXTENSION is used to extend
;;     the serializer to recognize implementation dependent objects.
;;
;;   READ-FROM-INPUT-SERIALIZER serializer                [function]
;;
;;     Read one object from the input serializer DESERIALIZER.
;;     It returns an eof object when it reaches the end of the input
;;     stream.
;;
;;   CALL-WITH-INPUT-SERIALIZER port proc &optional extension [function]
;;
;;     PROC must be a procedure which takes one argument, serializer.
;;     This procedure creates an input serializer from PORT and
;;     READER-EXTENSION argument, and passes it to PROC.
;;
;; Serialized format
;;
;;   Numbers, booleans, characters and the emptylist are written out the
;;   same way as the Scheme external presentation.
;;
;;   Other types are preceded by TAG presenting its type, then written
;;   out with its own way.   Those objects are also assigned a reference
;;   number in the order of appearance in the serializer, and if the same
;;   object (eq?-sense) appears more than once, the second and latter
;;   appearances are presented by REFERENCE, which is a reference tag and
;;   the number the object is assigned.
;;
;; Extending the serializer
;;
;;   You can extend the serializer to deal with implementation
;;   dependent objects, by passing a extention specification to
;;   the optional parameter EXTENTION to MAKE-{INPUT|OUTPUT}-SERIALIZER.
;;
;;   Extension specification is a list of reader/writer specification.
;;   A reader/writer specification is a list of four elements, namely
;;   a tag symbol, a test procedure, a writer procedure and a reader
;;   procedure.
;;
;;   When a serializer enconters an object of unknown type, it applies
;;   the test procedure on the object in the order it appears in the
;;   extension specification until it returns true.  Then the serializer
;;   writes a tag, and call a write procedure with a serializer, current
;;   output port, and the object to be written.   The reference is
;;   handled in the serializer so the write procedure need not care
;;   about it.   If not tests succeed, the serializer reports an error.
;;
;;   When a deserializer encounters unknown tag, it looks for the maching
;;   tag in the extension specification.  If found, the associated
;;   reader procedure is called with two arguments, an input serializer
;;   and the current input port.
;;   

;;;================================================================
;;; Implementation dependent part
;;;
;;; This module needs to have hashtable and error function.
;;;
;;; For hashtable, you can use SLIB's hashtable with most implementations.
;;; If your implementation has native hashtable sufpport, you can take
;;; advantage of it.  The hashtable's equality test must be done by eq?.
;;;
;;; Error function must take format-like string and parameters.  I think
;;; most Scheme implementation have it, too.
;;;
;;-----------------------------------------------------------------
;; If you use SLIB, uncomment the followings
;;
; (require 'hash-table)
; (define (pserializer:make-hash-table) (make-hash-table 13))
; (define pserializer:hash-table-put! (hash-associator eq?))
; (define pserializer:hash-table-get  (hash-inquirer eq?))
; (define pserializer:error error)
; (provide 'pserializer)

;;-----------------------------------------------------------------
;; If you use STk, uncomment the followings out
;;
(define pserializer:make-hash-table make-hash-table)
(define pserializer:hash-table-put! hash-table-put!)
(define (pserializer:hash-table-get tab key)  (hash-table-get tab key #f))
(define pserializer:error error)
(provide "pserializer")

;;;================================================================
;;; Implementation independent part
;;;

(define *pserializer:tag-symbol*    'y)
(define *pserializer:tag-pair*      'p)
(define *pserializer:tag-vector*    'v)
(define *pserializer:tag-string*    's)
(define *pserializer:tag-reference* 'r)

(define (pserializer:serializer->port serializer)
  (vector-ref serializer 0))
(define (pserializer:serializer->table serializer)
  (vector-ref serializer 1))
(define (pserializer:serializer->count serializer)
  (vector-ref serializer 2))
(define (pserializer:serializer->ext serializer)
  (vector-ref serializer 3))

(define (pserializer:incr-count! serializer)
  (vector-set! serializer 2 (+ (vector-ref serializer 2) 1)))

(define (pserializer:validate-extension proc e)
  (if (not (list? e))
      (error "~a: bad extension spec: ~s" proc e))
  (for-each (lambda (s)
              (if (not (and (list? s)
                            (= (length s) 4)
                            (symbol? (car s))
                            (procedure? (cadr s))
                            (procedure? (caddr s))
                            (procedure? (cadddr s))))
                  (error "~a: bad extension spec: ~s" proc s)))
            e)
  e)

;;--------------------------------------------------------------
;; Writer

(define (pserializer:make-output-serializer port . args)
  (let* ((extension
          (if (pair? args)
              (pserializer:validate-extension "make-output-serializer"
                                               (car args))
              '()))
         (ext-alist
          (map (lambda (e) `(,(cadr e) ,(car e) ,(caddr e))) extension))
         (serializer
          (vector port (pserializer:make-hash-table) 0 #f))
         (extension-proc
          (lambda (obj)
            (let loop ((ext ext-alist))
              (cond ((null? ext)
                     (error "write-to-output-serializer: object is unserializable: ~s"
                            obj))
                    ((or ((caar ext) obj)
                         (eq? (caar ext) #t))
                     (write (cadar ext) port) ; write tag
                     (display #\space port)
                     ((caddar ext) obj serializer))
                    (else (loop (cdr ext)))))))
         )
    (vector-set! serializer 3 extension-proc)
    serializer))

(define (pserializer:register-object-for-write obj serializer)
  (let* ((tab (pserializer:serializer->table serializer))
         (o (pserializer:hash-table-get tab obj)))
    (or o
        (let ((cnt (pserializer:serializer->count serializer)))
          (begin (pserializer:hash-table-put! tab obj cnt)
                 (pserializer:incr-count! serializer)
                 #f)))))

(define (pserializer:write-to-output-serializer obj serializer)
  (define port (pserializer:serializer->port serializer))
  (define (puttag tag)
    (write tag port) (display #\space port))
  (define (putobj obj)
    (write obj port) (display #\newline port))
    
  (if (or (eq? obj #f)
          (eq? obj #t)
          (eq? obj '())
          (number? obj)
          (char? obj))
      (putobj obj)
      (let ((ref (pserializer:register-object-for-write obj serializer)))
        (if ref
            (begin (puttag *pserializer:tag-reference*)
                   (putobj ref))
            (cond
             ((pair? obj)
              (puttag *pserializer:tag-pair*)
              (pserializer:write-to-output-serializer (car obj) serializer)
              (pserializer:write-to-output-serializer (cdr obj) serializer))
             ((symbol? obj)
              (puttag *pserializer:tag-symbol*)
              (putobj obj))
             ((string? obj)
              (puttag *pserializer:tag-string*)
              (putobj obj))
             ((vector? obj)
              (puttag *pserializer:tag-vector*)
              (putobj (vector-length obj))
              (for-each (lambda (elt)
                          (pserializer:write-to-output-serializer elt serializer))
                        (vector->list obj)))
             (else ((pserializer:serializer->ext serializer) obj))))
        )
      )
  )

(define (pserializer:call-with-output-serializer port proc . extension)
  (proc (apply pserializer:make-output-serializer port extension)))

;;--------------------------------------------------------------
;; Reader

(define (pserializer:make-input-serializer port . args)
  (let* ((extension (if (pair? args)
                        (pserializer:validate-extension
                         "read-from-input-serializer" (car args))
                        '()))
         (ext-hash (pserializer:make-hash-table)))
    (for-each (lambda (e)
                (pserializer:hash-table-put! ext-hash (car e) (cadddr e)))
              extension)
    (vector port (pserializer:make-hash-table) 0 ext-hash)))

(define (pserializer:register-object-for-read obj serializer . args)
  (let* ((tab (pserializer:serializer->table serializer))
         (cnt-provided? (and (pair? args) (integer? (car args))))
         (cnt (if cnt-provided?
                  (car args)
                  (pserializer:serializer->count serializer))))
    (pserializer:hash-table-put! tab cnt obj)
    (if (not cnt-provided?) (pserializer:incr-count! serializer))
    cnt))

(define (pserializer:lookup-object-for-read cnt serializer)
  (pserializer:hash-table-get (pserializer:serializer->table serializer) cnt))

(define (pserializer:dispatch-tag tag port serializer)
  (let ((proc (pserializer:hash-table-get
               (pserializer:serializer->ext serializer)
               tag)))
    (if proc
        (proc serializer)
        (pserializer:error "read-from-input-serializer: unknown tag: ~s" tag))))

(define (pserializer:error-premature-input)
  (pserializer:error "premature end of input stream"))

(define (pserializer:read-object port)
  (let ((obj (read port)))
    (if (eof-object? obj)
        (pserialzier:error-premature-input)
        obj)))

(define (pserializer:read-from-input-serializer serializer)
  (let* ((port (pserializer:serializer->port serializer))
         (tag (read port)))
    (cond ((eof-object? tag) tag)
          ((not (symbol? tag)) tag)
          ((eq? tag *pserializer:tag-reference*)
           (let* ((cnt (pserializer:read-object port))
                  (obj (pserializer:lookup-object-for-read cnt serializer)))
             (or obj
                 (pserializer:error "invalid reference number: ~s" cnt))))
          ((eq? tag *pserializer:tag-pair*)
           (let ((pair (cons #f #f)))
             (pserializer:register-object-for-read pair serializer)
             (let ((a (pserializer:read-from-input-serializer serializer))
                   (d (pserializer:read-from-input-serializer serializer)))
               (set-car! pair a)
               (set-cdr! pair d)
               pair)))
          ((eq? tag *pserializer:tag-symbol*)
           (let ((sym (pserializer:read-object port)))
             (pserializer:register-object-for-read sym serializer)
             sym))
          ((eq? tag *pserializer:tag-string*)
           (let ((str (pserializer:read-object port)))
             (pserializer:register-object-for-read str serializer)
             str))
          ((eq? tag *pserializer:tag-vector*)
           (let* ((len (pserializer:read-object port))
                  (vec (make-vector len)))
             (pserializer:register-object-for-read vec serializer)
             (let loop ((cnt 0))
               (if (< cnt len)
                   (let ((e (pserializer:read-from-input-serializer serializer)))
                     (vector-set! vec cnt e)
                     (loop (+ cnt 1)))
                   vec))))
          (else
           (pserializer:dispatch-tag tag port serializer)))
    )
  )

(define (pserializer:call-with-input-serializer port proc . extension)
  (proc (apply pserializer:make-input-serializer port extension)))

;;;================================================================
;;; Export procedures
;;;

(define make-output-serializer pserializer:make-output-serializer)
(define write-to-output-serializer pserializer:write-to-output-serializer)
(define call-with-output-serializer pserializer:call-with-output-serializer)

(define make-input-serializer pserializer:make-input-serializer)
(define read-from-input-serializer pserializer:read-from-input-serializer)
(define call-with-input-serializer pserializer:call-with-input-serializer)

(define register-object-to-input-serializer
  pserializer:register-object-for-read)

(define serializer->port pserializer:serializer->port)
